==> symbol-processing-aa-embeddable-2.1.0-1.0.29-sources.jar: org/jetbrains/kotlin/fir/analysis/checkers/extra/FirAnonymousUnusedParamChecker.kt <==
/*
 * Copyright 2010-2024 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.fir.analysis.checkers.extra

import ksp.org.jetbrains.kotlin.KtFakeSourceElementKind
import ksp.org.jetbrains.kotlin.diagnostics.DiagnosticReporter
import ksp.org.jetbrains.kotlin.diagnostics.reportOn
import ksp.org.jetbrains.kotlin.fir.FirElement
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.MppCheckerKind
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.context.CheckerContext
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.declaration.FirAnonymousFunctionChecker
import ksp.org.jetbrains.kotlin.fir.analysis.diagnostics.FirErrors.UNUSED_ANONYMOUS_PARAMETER
import ksp.org.jetbrains.kotlin.fir.declarations.FirAnonymousFunction
import ksp.org.jetbrains.kotlin.fir.references.FirResolvedNamedReference
import ksp.org.jetbrains.kotlin.fir.symbols.impl.FirValueParameterSymbol
import ksp.org.jetbrains.kotlin.fir.visitors.FirVisitor
import ksp.org.jetbrains.kotlin.name.SpecialNames

object FirAnonymousUnusedParamChecker : FirAnonymousFunctionChecker(MppCheckerKind.Common) {
    override fun check(declaration: FirAnonymousFunction, context: CheckerContext, reporter: DiagnosticReporter) {
        context.lambdaBodyContext?.checkUnusedParams(declaration, context, reporter)
    }

    class LambdaBodyContext(private val outermostLambda: FirAnonymousFunction) {
        internal fun checkUnusedParams(declaration: FirAnonymousFunction, context: CheckerContext, reporter: DiagnosticReporter) {
            // We need to check only outermost Lambda which will detect unused params of nested Lambdas too.
            if (declaration != outermostLambda)
                return

            val unusedParams = declaration.getReportableParameters()

            declaration.body?.accept(unusedParamsVisitor, unusedParams)

            unusedParams.forEach {
                reporter.reportOn(it.source, UNUSED_ANONYMOUS_PARAMETER, it, context)
            }
        }
    }

    private fun FirAnonymousFunction.getReportableParameters(): MutableSet<FirValueParameterSymbol> {
        return valueParameters
            .filter { it.source?.kind !is KtFakeSourceElementKind && it.name != SpecialNames.UNDERSCORE_FOR_UNUSED_VAR }
            .mapTo(mutableSetOf()) { it.symbol }
    }

    private val unusedParamsVisitor: FirVisitor<Unit, MutableSet<FirValueParameterSymbol>> =
        object : FirVisitor<Unit, MutableSet<FirValueParameterSymbol>>() {
            override fun visitElement(element: FirElement, data: MutableSet<FirValueParameterSymbol>) {
                if (data.isNotEmpty()) {
                    element.acceptChildren(this, data)
                }
            }

            override fun visitResolvedNamedReference(
                resolvedNamedReference: FirResolvedNamedReference,
                data: MutableSet<FirValueParameterSymbol>,
            ) {
                data.remove(resolvedNamedReference.resolvedSymbol)
            }

            override fun visitAnonymousFunction(anonymousFunction: FirAnonymousFunction, data: MutableSet<FirValueParameterSymbol>) {
                if (!anonymousFunction.isLambda) {
                    this.visitElement(anonymousFunction, data)
                    return
                }

                data.addAll(anonymousFunction.getReportableParameters())

                anonymousFunction.acceptChildren(this, data)
            }
        }
}

fun createLambdaBodyContext(lambda: FirAnonymousFunction, context: CheckerContext): FirAnonymousUnusedParamChecker.LambdaBodyContext {
    return context.lambdaBodyContext ?: FirAnonymousUnusedParamChecker.LambdaBodyContext(lambda)
}

==> symbol-processing-aa-embeddable-2.2.0-2.0.2-sources.jar: org/jetbrains/kotlin/fir/analysis/checkers/extra/FirAnonymousUnusedParamChecker.kt <==
/*
 * Copyright 2010-2024 JetBrains s.r.o. and Kotlin Programming Language contributors.
 * Use of this source code is governed by the Apache 2.0 license that can be found in the license/LICENSE.txt file.
 */

package ksp.org.jetbrains.kotlin.fir.analysis.checkers.extra

import ksp.org.jetbrains.kotlin.KtFakeSourceElementKind
import ksp.org.jetbrains.kotlin.diagnostics.DiagnosticReporter
import ksp.org.jetbrains.kotlin.diagnostics.reportOn
import ksp.org.jetbrains.kotlin.fir.FirElement
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.MppCheckerKind
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.context.CheckerContext
import ksp.org.jetbrains.kotlin.fir.analysis.checkers.declaration.FirAnonymousFunctionChecker
import ksp.org.jetbrains.kotlin.fir.analysis.diagnostics.FirErrors.UNUSED_ANONYMOUS_PARAMETER
import ksp.org.jetbrains.kotlin.fir.declarations.FirAnonymousFunction
import ksp.org.jetbrains.kotlin.fir.references.FirResolvedNamedReference
import ksp.org.jetbrains.kotlin.fir.symbols.impl.FirValueParameterSymbol
import ksp.org.jetbrains.kotlin.fir.visitors.FirVisitor
import ksp.org.jetbrains.kotlin.name.SpecialNames

object FirAnonymousUnusedParamChecker : FirAnonymousFunctionChecker(MppCheckerKind.Common) {
    context(context: CheckerContext, reporter: DiagnosticReporter)
    override fun check(declaration: FirAnonymousFunction) {
        context.lambdaBodyContext?.checkUnusedParams(declaration, context, reporter)
    }

    class LambdaBodyContext(private val outermostLambda: FirAnonymousFunction) {
        internal fun checkUnusedParams(declaration: FirAnonymousFunction, context: CheckerContext, reporter: DiagnosticReporter) {
            // We need to check only outermost Lambda which will detect unused params of nested Lambdas too.
            if (declaration != outermostLambda)
                return

            val unusedParams = declaration.getReportableParameters()

            declaration.body?.accept(unusedParamsVisitor, unusedParams)

            unusedParams.forEach {
                reporter.reportOn(it.source, UNUSED_ANONYMOUS_PARAMETER, it, context)
            }
        }
    }

    private fun FirAnonymousFunction.getReportableParameters(): MutableSet<FirValueParameterSymbol> {
        return valueParameters
            .filter { it.source?.kind !is KtFakeSourceElementKind && it.name != SpecialNames.UNDERSCORE_FOR_UNUSED_VAR }
            .mapTo(mutableSetOf()) { it.symbol }
    }

    private val unusedParamsVisitor: FirVisitor<Unit, MutableSet<FirValueParameterSymbol>> =
        object : FirVisitor<Unit, MutableSet<FirValueParameterSymbol>>() {
            override fun visitElement(element: FirElement, data: MutableSet<FirValueParameterSymbol>) {
                if (data.isNotEmpty()) {
                    element.acceptChildren(this, data)
                }
            }

            override fun visitResolvedNamedReference(
                resolvedNamedReference: FirResolvedNamedReference,
                data: MutableSet<FirValueParameterSymbol>,
            ) {
                data.remove(resolvedNamedReference.resolvedSymbol)
            }

            override fun visitAnonymousFunction(anonymousFunction: FirAnonymousFunction, data: MutableSet<FirValueParameterSymbol>) {
                if (!anonymousFunction.isLambda) {
                    this.visitElement(anonymousFunction, data)
                    return
                }

                data.addAll(anonymousFunction.getReportableParameters())

                anonymousFunction.acceptChildren(this, data)
            }
        }
}

fun createLambdaBodyContext(lambda: FirAnonymousFunction, context: CheckerContext): FirAnonymousUnusedParamChecker.LambdaBodyContext {
    return context.lambdaBodyContext ?: FirAnonymousUnusedParamChecker.LambdaBodyContext(lambda)
}
